'Filename   RS485_Master.bas
'Version    1.00
'Date       2016-09-01
'Purpose    Test communication ATU Controller (master) with TRX-COM (slave)
'           Requests perodically RS485 slave for the TRX frequency
'           TRX frequency wrdTRXFreq is sent in 2 bytes bytTRXFreq
'           bytTRXFreq(2) are overlayed with wrdTRXFreq
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega32
'           Crystal 16 MHz
'           Port B0...B5: LCD
'Software   RS485 protocoll
'           Byte  Content
'           1     Address of message destination ("to")
'           2     Address of message source  ("from")
'           3     Command
'           4     Data byte 1
'           5     Data byte 2
'           6     Data byte 3
'           7     Checksum of bytes 1 to 6 (CRC8, 1 byte)
'Compiler   BASCOM-AVR 2.0.7.6
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor
'           Flash used: 1,808 bytes (5%)

'==============================================================================

'FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'In ATMEGA-Data sheets: Bit = 1 -> not set
'                       Bit = 0 -> set
'OCDEN      [ ]
'JTAGEN     [ ] (disabled)
'SPIEN      [x] Do not change!
'CKOPT      [ ]
'EESAVE     [ ]
'BOOTSZ     Boot Flash size=256 words (minimal boot size)
'BOOTRST    [ ]
'BODLEVEL   Brown-out detection at VCC=4.0V
'BODEN      [ ]
'SUT_CKSEL  Ext. Crystal/Resonator High Freq.; Start-up time 16k + 65ms

'Hex fuse values: High: DF, Low: 7F

$regfile = "m32def.dat"                           'ATMega32
$crystal = 16000000                               '16 MHz, baud error 0.16%, works
'$crystal = 14745600                               '14.7456 MHz xtal, baud error 0%
$baud = 9600                                      'Baud rate RS485
$hwstack = 40
$swstack = 40
$framesize = 40

'Configure port A (not used)---------------------------------------------------
DdrA = &B00000000                                 'PortA = input
PortA = &B11111111                                'Pullup

'Configure LCD on Port B ------------------------------------------------------
Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortB.5 , Db6 = PortB.4 , Db5 = PortB.3
Config Lcdpin = Pin , Db4 = PortB.2 , E = PortB.1 , Rs = PortB.0
Config Lcdbus = 4

'Configure port C (not used)---------------------------------------------------
DDRC = &B0000000                                  'PortC = input
PortC = &B1111111                                 'Pullup

'Configure port D -------------------------------------------------------------
DDRD = &B00000100                                 'D.0   RS485 RXD
'                                                 'D.1   RS485 TXD
'                                                 'D.2   RS485 RE/DE output
PORTD = &B11111000                                'Pullup D.3 to D.7
bitRS485 Alias Portd.2                            '=1:RS485 Transmit, =0: Receive

'Variables Program control ----------------------------------------------------
Const On = 1
Const Off = 0

'Variables RS485-Communication ------------------------------------------------
Const bytConAdr = &H00                            'ATU Controller  COM address
Const bytATUAdr = &H01                            'ATU remote unit COM address
Const bytTRXAdr = &H02                            'TRX Com         COM address
Const bytFreqCall = &HFA                          'Hex code for TRX freq call & response
Const bytDummy = &HFF                             'Dummy byte
Const bytFrameLen = 7                             'Length of RS485 frame
Const bytFrameLenMinus1 = bytFrameLen - 1         '... minus 1
Dim bitComBusy As Bit                             '=1: RS485 communication active
Dim bytComErr As Byte                             '0: Message read is OK, >0 else
Dim bytCRC8 As Byte                               'Crc8 checksum (1 byte)
Dim bytBuffCount As Byte                          'Buffer byte counter
Dim bytBuffInp(bytFrameLen) As Byte               'Input Buffer
Dim bytBuffOut(bytFrameLen) As Byte               'Output Buffer
Dim wrdTRXFreq As Word                            'TRX frequency (kHz)
Dim bytTRXFreq(2) As Byte At wrdTRXFreq Overlay

Const ComDebug = 1                                '=1: show LCD debug information

'Common Variables, multiply used ----------------------------------------------
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte

'Configure Timer0-Interrupt (Timer0 = 8bit, 2^8=256 counts) -------------------
'Configuration with xtal 16MHz for use with a frequency counter
'Overflow time: Overflow counts * Prescale / crystal frequency
'Overflow counts = 256 - PresetTimer0 = 200 (PresetTimer0 = 56)
'= 200 * 1024 / 16.000.000 = 12.8 msec = ~ 78 Hz (crystal 16 MHz)
'Possible Prescales: 8, 64, 256, 1024, here 1024
Config Timer0 = Timer , Prescale = 1024
Const PresetTimer0 = 56                           'Initialize Timer0 (shorten time, >0)
On Timer0 Timer0_isr                              'On Overflow go to Timer0_isr
Dim bytCallTRX_cnt As Byte                        'Timer counter for TRX call
Const bytCallTRX_runs = 35                        'Timer counts for TRX call ~0.5sec
Dim bytTimeout_cnt As Byte                        'Timer counter for time out
Const bytTimeout_runs = 70                        'Timer counts for timeout ~1sec

'Declare subs & functions -----------------------------------------------------
Declare Sub GetTRXFreq

'Start timer0, initialize data ------------------------------------------------
Enable Timer0
Timer0 = Presettimer0                             'initialize Timer0
SREG.7 = 1                                        'Enable all interrupts

bytCallTRX_cnt = 0                                'Initialize counter for TRX call
bytTimeout_cnt = 0                                'Initialize counter for time out
bytBuffCount = 0                                  'Initialize RS485 Input buffer counter
bitRS485 = Off                                    'RS485 receive
bitComBusy = 0                                    'RS485 communication not busy

Cls
Cursor Off
Locate 1 , 1
LCD "RS485 Master 100"
Wait 1
Cls

'Check TRX communication unit -------------------------------------------------
Locate 1 , 1
Lcd "TRX Com OK?"
Locate 2 , 1
bytTmp1 = 0
For bytTmp0 = 1 To 3                              'Try it 3 times
    Call GetTRXFreq                               'Send frequency request to TRX
    If bytComErr = 0 Then                         'Freq received, TRX unit is ready
       bytTmp1 = 1
       Exit For
    End If
    Waitms 100
Next bytTmp0
If bytTmp1 = 1 Then
   Lcd "... is OK!"
Else
   Lcd "... is not OK!"
   Stop                                           'Program stopped
End If
Wait 1
Cls

Do

   'Check wait time for TRX call
   If bitComBusy = 0 Then                         'RS485 bus is not busy
      If bytCallTRX_cnt > bytCallTRX_runs Then    'Wait time elapsed
         Call GetTRXFreq                          'Send TRX frequency request
         bytCallTRX_cnt = 0                       'Reset Call timer counter

         'Insert bytComErr error handling here, bytComErr=0: all OK

      End If
   End If

   'Communication with second bus slave
   If bitComBusy = 0 Then                         'RS485 bus is not busy
      'Call second slave...
   End If

Loop

'=======================================================================Timer0=
Timer0_isr:
'Interrupt Service Routine for Timer0 (Time out & TRX call scheduler)

   Timer0 = Presettimer0

   'Communication timeout -----------------------------------------------------
   Incr bytTimeout_cnt                            'Counter for receive time out

   'TRX frequency request scheduler -------------------------------------------
   Incr bytCallTRX_cnt

Return

'==============================================================================
Sub GetTRXFreq
'Call frequency (kHz) from TRX (slave)
'Byte  Content
'1     Address of message destination ("to")
'2     Address of message source  ("from")
'3     Command
'4     Data byte 1
'5     Data byte 2
'6     Data byte 3
'7     Checksum of bytes 1 to 7 (CRC8, 1 byte)

'"PrintBin" is used (no ASCII transformation)
'Output:
'wrdTRXFreq    Frequency (kHz) received from TRX, overlayed with bytTRXfreq
'bytComErr     = 0: OK, frequency received
'              = 1: Time out, Input Buffer bytBuffInp not complete
'              = 2: Message destination is not me, the controller
'              = 3: Message source is not the TRX
'              = 4: Checksum is not OK
'              = 5: No frequency received
'              = 6: Command not OK

   bitComBusy = 1                                 'RS485 communication busy
   bitRS485 = On                                  'set MAX485 to transmit
   Waitms 2                                       'Wait for MAX485 to switch
   UCSRA.6 = 1                                    'Clear UART Transmit Complete bit
   'Send TRX frequency request, configure protocoll buffer
   bytBuffOut(1) = bytTRXAdr                      'destination address (TRX)
   bytBuffOut(2) = bytConAdr                      'Sender address (ATU controller)
   bytBuffOut(3) = bytFreqCall                    'Request TRX frequency
   bytBuffOut(4) = bytDummy                       'not used, dummy
   bytBuffOut(5) = bytDummy                       'not used, dummy
   bytBuffOut(6) = bytDummy                       'not used, dummy
   bytCRC8 = Crc8(bytBuffOut(1) , bytFrameLenMinus1)
   bytBuffOut(7) = bytCRC8                        'CRC8
   Printbin bytBuffOut(1) ; bytFrameLen           'Send array bytBuffOut
   Do
   Loop Until UCSRA.6 = 1                         'Wait until all bytes are sent
   bitRS485 = Off                                 'Set MAX485 to receive
   Waitms 2                                       'Wait for MAX485 to switch

   #If ComDebug = 1
      Locate 1 , 1
      Lcd bytBuffOut(1) ; " " ; bytBuffOut(2) ; " " ; bytBuffOut(3) ; " " ; bytBuffOut(7)
      Locate 1 , 16
      LCD "*"
   #EndIf

   bytTimeout_cnt = 0                             'Start Receive time out (Timer0)
   bytBuffCount = 0                               'Initialize input buffer counter

   Do
      If Ischarwaiting() = 1 Then                 'Wait for character
         Inputbin bytTmp0                         'Read character
         Incr bytBuffCount                        'Next address in buffer
         bytBuffInp(bytBuffCount) = bytTmp0       'Save character in buffer
      End If
   Loop Until bytBuffCount = bytFrameLen Or bytTimeout_cnt > bytTimeout_runs

   If bytBuffCount = bytFrameLen Then             'Response is complete
      If bytBuffInp(1) = bytConAdr Then           'It's me, the ATU controller
         bytCRC8 = Crc8(bytBuffInp(1) , bytFrameLenMinus1)
         If bytCRC8 = bytBuffInp(7) Then          'Checksum is OK
            If bytBuffInp(2) = bytTRXAdr Then     'It's from the TRX
               If bytBuffInp(3) = bytFreqCall Then       'It's frequency request
                  If bytBuffInp(4) = 0 Then       'Error code 0: OK, freq is sent
                     bytTRXFreq(1) = bytBuffInp(5)
                     bytTRXFreq(2) = bytBuffInp(6)
                     bytComErr = 0                'Frequency is received
                  Else
                     wrdTRXFreq = 0
                     bytComErr = 5                'No freq sent
                  End If
               Else
                  bytComErr = 6                   'command not OK
               End If
            Else
               bytComErr = 3                      'Source address not OK
            End If
         Else
            bytComErr = 4                         'Checksum not OK
         End If
      Else
         bytComErr = 2                            'destination address not OK
      End If
   Else                                           'Response not complete or missing
      bytComErr = 1                               'Timeout
   End If

   bitComBusy = 0                                 'RS485 communication not busy
   Waitms 10

   #If ComDebug = 1
      Locate 2 , 1
      Lcd bytBuffInp(1) ; " " ; bytBuffInp(2) ; " " ; bytBuffInp(4) ; " " ; wrdTRXFreq ; " " ; bytBuffCount ; " E" ; bytComErr
      Locate 2 , 16                               'Show scheduler heart beat
      LCD "*"
      Waitms 500
      Locate 2 , 16
      LCD " "
      Locate 1 , 16
      LCD " "
   #EndIf

End Sub

End