'Filename   RS485_Slave.bas
'Version    1.00
'Date       2016-09-11
'Purpose    RS485 communication (slave)
'           RS485 master (ATU controller) requests the TRX frequency
'           Slave response is the current frequency (kHz)
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega48, Standard HD44780 16x2 LCD for debugging
'Software   RS485 protocoll
'           Byte  Content
'           1     Address of message destination ("to")
'           2     Address of message source  ("from")
'           3     Command (request) or error code (response)
'           4     Data byte 1
'           5     Data byte 2
'           6     Data byte 3
'           7     Checksum of bytes 1 to 6 (CRC8, 1 byte)
'Compiler   BASCOM-AVR 2.0.7.6
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor.
'           Flash used: 790 bytes (19%)
'Fuse bits  can be set with AVR Studio 4.0 & ISP-PROG via USB
'           FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'           In ATMEGA-Data sheets:  Bit = 1 -> not set
'                                   Bit = 0 -> set
'           SELFPRGEN  [ ]
'           RSTDISBL   [ ]
'           DWEN       [ ]
'           SPIEN      [x]
'           WDTON      [ ]
'           EESAVE     [ ]
'           BODLEVEL   Brown-out detection at VCC=2.7V
'           CKDIV8     [ ]
'           CKOUT      [ ]
'           SUT_CKSEL  Ext. Crystal Osc. High Freq., Start-up time 16k CK/14 + 64ms

'Edit history

$regfile = "m48def.dat"                           'Controller ATmega48
$crystal = 14745600                               '14.7456 MHz xtal
$baud = 9600                                      'Baud rate RS485
$hwstack = 60
$swstack = 40
$framesize = 40

'Configure Port C (not used) --------------------------------------------------
DDRC = &B00000000                                 'PortC = input
PORTC = &B11111111                                'Pullup PC1...PC7

'Configure LCD at  Ports B & D ------------------------------------------------
'LCD-DB7 PD.7
'LCD-DB6 PB.0
'LCD-DB5 PD.6
'LCD-DB4 PB.1
'LCD-E   PD.5
'LCD-RS  PB.2
Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortD.7 , Db6 = PortB.0 , Db5 = PortD.6
Config Lcdpin = Pin , Db4 = PortB.1 , E = PortD.5 , Rs = PortB.2
Config Lcdbus = 4

'RS485 communication at Port D.0...D.2, D.0=RXD, D.1=TXD
DDRD.2 = 1                                        'D.2 RS485 RE/DE is output
bitRS485 Alias Portd.2                            '=1:Transmit, =0: Receive
DDRD.3 = 0                                        'Port D3=Input
DDRD.4 = 0                                        'Port D4=Input
PORTD.3 = 1                                       'Port D3 pullup
PORTD.4 = 1                                       'Port D4 pullup

'Variables Program control ----------------------------------------------------
Const On = 1
Const Off = 0

'Variables RS485-Communication ------------------------------------------------
Const bytConAdr = &H00                            'ATU Controller  COM address
Const bytATUAdr = &H01                            'ATU remote unit COM address
Const bytTRXAdr = &H02                            'TRX             COM address
Const bytDummy = &HFF                             'Dummy byte
Const bytFreqCall = &HFA                          'Hex code for TRX freq call & response
Const bytFrameLen = 7                             'Length of RS485 frame
Const bytFrameLenMinus1 = bytFrameLen - 1         '... minus 1
Dim bytBuffCount As Byte                          'Byte Buffer counter
Dim bytComErr As Byte
Dim bytCRC8 As Byte                               'Crc8 checksum (1 byte)
Dim bytBuffInp(bytFrameLen) As Byte               'Input Buffer
Dim bytBuffOut(bytFrameLen) As Byte               'Output Buffer
Dim bytFreqTest As Byte
Dim bytTRXSend As Byte                            ':1 Message OK, send response to ATU
Dim wrdTRXFreq As Word
Dim bytTRXFreq(2) As Byte At wrdTRXFreq Overlay

Const ComDebug = 0                                '=1: show LCD debug information

'Common variables, multiply used ----------------------------------------------
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte

'Subs and functions -----------------------------------------------------------
Declare Sub SndTRXFreq

'Initialize data --------------------------------------------------------------
bitRS485 = OFF                                    'RS485 receive
wrdTRXFreq = 3500
bytFreqTest = 0
bytBuffCount = 0                                  'Com buffer counter

'Enable UART interrupt service routine ----------------------------------------
On URXC RXD_isr                                   'Go to RXD_isr if character is received
Enable URXC                                       'Enable UART receive interrupt
SREG.7 = 1                                        'Enable all interrupts

'Splash screen ----------------------------------------------------------------
#If ComDebug = 1
   Cls
   Cursor Off
   Upperline
   LCD "RS485 Slave"
   Wait 1
   Cls
#EndIf

Do

   If bytBuffCount = bytFrameLen Then             'Master request is complete
      'Don't know why, Waitms must be >= 15ms before setting bytBuffCount = 0 ???
      'Without or shorter less than bytFrameLen bytes will be sent to master
      'So master will encounter timeout
      Waitms 15
      bytBuffCount = 0
      Call SndTRXFreq                             'Check request & send frequency
      If bytFreqTest < 30 Then
         wrdTRXFreq = wrdTRXFreq + 10             'Loop 3.5...3.8MHz
         Incr bytFreqTest
      Else
         wrdTRXFreq = 3500
         bytFreqTest = 0
      End If
   End If

Loop

'==============================================================================
'URXC UART receive complete interrupt service routine
'Read current received Byte from UART register UDR
'Output:  bytBuffInp(1...bytBuffCount), Input buffer
'         bytBuffCount, current length of input buffer

RXD_isr:
   Incr bytBuffCount
   bytBuffInp(bytBuffCount) = UDR                 'Put next byte into buffer
Return

'==============================================================================
Sub SndTRXFreq
'Get frequency request (kHz) from ATU
'If it's OK, send frequency back to ATU
'Byte  Content
'1     Address of message destination ("to")
'2     Address of message source ("from")
'3     Command for destination or error code returned from destination resp.
'4     Data byte 1
'5     Data byte 2
'6     Data byte 3
'7     Checksum of bytes 1 to 6 (CRC8, 1 byte)
'Output:
'bytComErr  = 0: Message received is OK
'           = 1: Command not OK
'           = 2: Checksum not OK
'           = 3: Source address not OK, is not from the master
'           = 4: Destination address not OK, is not for me, the slave

   #If ComDebug = 1
      Locate 1 , 1
      Lcd bytBuffInp(1) ; " " ; bytBuffInp(2) ; " " ; bytBuffInp(3) ; " " ; bytBuffInp(7)
   #EndIf


   'Check the received message:
   '- is the destination my address?
   '- is the source the ATU controller?
   '- is the received checksum OK?
   '- is the 'Send frequency' command OK?
   bytTRXSend = 1                                 'Send response
   If bytBuffInp(1) = bytTRXAdr Then              'It's for me, the TRX
      If bytBuffInp(2) = bytConAdr Then           'It's from the ATU controller
         bytTmp1 = Crc8(bytBuffInp(1) , bytFrameLenMinus1)
         If bytTmp1 = bytBuffInp(7) Then          'Checksum is OK
            If bytBuffInp(3) = bytFreqCall Then   'It's frequency request
               bytComErr = 0                      'Message is OK
            Else
               bytComErr = 1                      'No frequency request
            End If
         Else
            bytComErr = 2                         'Checksum not OK
         End If
      Else
         bytComErr = 3                            'Source is not ATU controller
         bytTRXSend = 0                           'Do not send response
      End If
   Else
      bytComErr = 4                               'Message was not for me
      bytTRXSend = 0                              'Do not send response
   End If

   If bytTRXSend = 1 Then                         'Send TRX frequency response
      bitRS485 = On                               'set MAX485 to transmit
      Waitms 2                                    'Wait for MAX485 to switch
      UCSR0A.6 = 1                                'Clear UART Transmit Complete bit
      bytBuffOut(1) = bytConAdr                   'Destination address (ATU controller)
      bytBuffOut(2) = bytTRXAdr                   'Source   address (TRX)
      bytBuffOut(3) = bytFreqCall                 'TRX frequency call command
      bytBuffOut(4) = bytComErr                   'Com error code
      If bytComErr = 0 Then                       'No Com error
         bytBuffOut(5) = bytTRXFreq(1)            'Frequency high byte
         bytBuffOut(6) = bytTRXFreq(2)            'Frequency low  byte
      Else
         bytBuffOut(5) = bytDummy                 'not used, dummy
         bytBuffOut(6) = bytDummy                 'not used, dummy
      End If
      bytTmp1 = Crc8(bytBuffOut(1) , bytFrameLenMinus1)
      bytBuffOut(7) = bytTmp1                     'CRC8
      Printbin bytBuffOut(1) ; bytFrameLen        'Send array bytBuffOut
      Do
      Loop Until UCSR0A.6 = 1                     'Wait until all bytes are sent
      bitRS485 = OFF                              'ready for receive again
      Waitms 2                                    'Wait for MAX485 to switch

      #If ComDebug = 1
         Locate 2 , 1
         LCD bytBuffOut(1) ; " " ; bytBuffOut(2) ; " " ; bytBuffOut(3) ; " " ; bytBuffOut(4) ; " " ; wrdTRXFreq
      #EndIf

   End If

End Sub

'==============================================================================
End